"""Parses NODC code files."""

import csv
import collections
from os.path import join
import re

import pandas as pd


def _is_table(line):
    if re.match(' \d{4}   \w', line):
        return True
    else:
        return False


def _is_code(line):
    if re.match(' [\w\s]{6} -- [\w\.]', line):
        return True
    else:
        return False


def parse_code_file(code_file):
    """Parse a code file into code tables.

    A code file has several code tables described within and possibly a header.
    Below is an example:


    Code Tables Used with Intertidal Organisms
    and Habitats File type (F030)                                               


     0008   COMPOSITION 
     ----   --------------------
          1 -- BEDROCK  
          2 -- RUBBLE GREATER THAN 2 SQUARE FEET
          3 -- RUBBLE LESS THAN OR EQUAL TO 2 SQUARE FEET   
          4 -- COMBINATION OF 2 AND 3 ABOVE 
          5 -- GRAVEL   
          6 -- SAND 
          7 -- MUD  

     0009   COVER (030) 
     ----   --------------------
          1 -- BARE 
          2 -- LIGHT
          3 -- MEDIUM   
          4 -- HEAVY

    Args:
        code_file: Full path to code file.

    Returns:
        OrderedDict of code tables indexed by table number. Each code table is
        a dict with name (str), number (str), and codes (OrderedDict). The
        codes OrderedDict is keyed by code and has code descriptions (str) as
        values.
    """
    code_tables = collections.OrderedDict()
    with open(code_file) as f:
        lines = f.read().splitlines()
    current_codes = None
    current_code = None
    for line in lines:
        if line == '':  # not in a table
            current_codes = None
            current_code = None
        elif line[0] != ' ':  # possible line continuation
            if current_code is not None:
                current_codes[current_code] += ' ' + line
            else:
                pass  # NODC TAXONOMIC CODE
        elif _is_table(line):
            number, name = line[1:5], line[8:]
            current_table = {'name': name, 'number': number}
            current_codes = collections.OrderedDict()
            current_table['codes'] = current_codes
            code_tables[number] = current_table
        elif _is_code(line) and current_codes is not None:
            code, description = line[:7].strip(), line[11:].strip()
            current_codes[code] = description
            current_code = code
        else:
            pass  # possible metadata
    return code_tables


def codes_to_dataframes(code_file):
    """Parse a code file into pandas DataFrames.

    A code file has several code tables described within, and each table has
    a unique table number.

    Args:
        code_file: Full path to code file.

    Returns:
        dict keyed by code number of pandas DataFrames. Each DataFrame has
        Code and Code_Meaning columns, and a name property indicating the
        code table name.
    """
    dataframes = {}
    columns = ['Code', 'Code_Meaning']
    code_tables = parse_code_file(code_file)
    for number, table in code_tables.iteritems():
        rows = [[code, desc] for code, desc in table['codes'].iteritems()]
        df = pd.DataFrame(rows, columns=columns)
        df.name = table['name']
        dataframes[number] = df
    return dataframes
