"""Handles record formats."""

import csv
from datetime import datetime
import os



def _read_record_format(record_format_filename):
    """Reads a record format file.

    The record format file is a CSV file that describes each field in a
    column based NODC archive data file.  The record format file has these
    columns:
        Field: Field name.
        Description: Description of the field.
        Start_Column: 1-based start column for the field.
        Field_Length: Number of characters in the field.
        Code (Optional): NODC code table number for code fields.
        Date_Format (Optional): Python strftime format for date fields.
        Strip (Optional): y if whitespace should be stripped from content.
        Number_Format (Optional): For numeric fields, use i for integer and
            .f# for decimals, e.g, .f3 for a number with three decimal places.
        Concatenate (Optional): If field is a concatenation of other fields,
            use a comma-separated list of field names here, e.g.,
            NODC_TRACK_NUMBER,STATION_NUMBER

    Args:
        record_format_filename: Record format filename.

    Returns:
        List with one dictionary for each row in the record format file, where
        each dictionary's members represent each column from the format file.
    """
    cells = []
    with open(record_format_filename) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            code = row['Code']
            if code != '':
                row['Code'] = code.zfill(4)  # Excel strips leading zeroes
            cells.append(row)
    return cells


def read_record_formats(folder):
    """Reads file formats for record types from CSV files.

    Reads file formats for record types from CSV files. Each file should be
    named [record number]_[record type], e.g., '1_File Header.csv'.

    Args:
        folder: String indicating folder containing record format files.

    Returns:
        Dictionary of record formats indexed by record type.
    """
    ignore = ['join_by_sequence.csv', 'joins.csv']
    files = [os.path.join(folder, f) for f in os.listdir(folder)
             if f.endswith('.csv') and f not in ignore]
    formats = {}
    for f in files:
        parts =  os.path.basename(f).split('_', 1)
        record_number = parts[0]
        record_name =  os.path.splitext(parts[1])[0]
        formats[record_number] = {'fields': _read_record_format(f),
                                  'name': record_name}        
    return formats


def _slice(line, start_column, field_length):
    """Extracts a slice from a line of text with a one-based index.

    Slices the input line of text to extract a portion defined by a start
    column and the length of the cell.  We assume NODC column definitions,
    which are one-based, e.g., the first column is column 1.

    Args:
        line: The string from which a portion is to be extracted.
        start_column: One-based start index of the cell. 
        field_length: Number of characters in the cell.

    Returns:
        The extracted string.
    """
    python_start_index = int(start_column) - 1
    python_end_index = python_start_index + int(field_length)
    return line[python_start_index:python_end_index]


def _parse_degrees(data, num_fmt):
    """Extracts a slice from a line of text with a one-based index.

    Slices the input line of text to extract a portion defined by a start
    column and the length of the cell.  We assume NODC column definitions,
    which are one-based, e.g., the first column is column 1.

    Args:
        data: The string representing the degrees, e.g., 753433N.
        num_fmt: Degree number format. Allowable values are:
            DDMMXXN, DDDMMXXE, DDMMSSN, DDDMMSSE
            where:
                D = degrees
                M = minutes
                X = decimal minutes
                S = seconds
                N = data will indicate N (north) or S (south)
                E = data will indicate E (east) or W (west)

    Returns:
        Degrees as float.
    """
    need_minus = ['W', 'S']
    if num_fmt == 'DDMMXXN':
        deg = int(data[:2]) + int(data[2:4]) / 60. + int(data[4:6]) / 6000.
    elif num_fmt == 'DDDMMXXE':
        deg = int(data[:3]) + int(data[3:5]) / 60. + int(data[5:7]) / 6000.
    elif num_fmt == 'DDMMSSN':
        deg = int(data[:2]) + int(data[2:4]) / 60. + int(data[4:6]) / 3600.
    elif num_fmt == 'DDDMMSSE':
        deg = int(data[:3]) + int(data[3:5]) / 60. + int(data[5:7]) / 3600.
    if data[-1] in need_minus:
        deg = deg * -1
    return deg


def _concat(expression, data_row, record_format):
    """Concatenates two or more fields with periods.

    Concatenates two or more fields with periods.  Fields participating in the
    concatenation must have already been processed and populated in the input
    data row.

    Args:
        expression: Comma-separated list of field names, e.g.,
            NODC_TRACK_NUMBER,STATION_NUMBER
        data_row: List of content from other fields in the same order as each
            field appears in the record_format.
        record_format: List with one dictionary for each row in the record
            format file, where each dictionary's members represent each column
            from the format file.  The only column used by this function is the
            Field column, which provides the field name.

    Returns:
        Period-concatenated content of two or more fields.
    """
    fields = [f.strip() for f in expression.split(',')]
    parts = []
    for field in fields:
        for index, cell in enumerate(record_format):
            if cell['Field'] == field:
                part = data_row[index]
                if part is None:
                    return None
                parts.append(str(part))
    return '.'.join(parts)


def _parse_line(line, record_format):
    """Parses a line from a data file.

    Args:
        line: The column-based line of text to parse
        record_format: List with one dictionary for each row in the record
            format file, where each dictionary's members represent each column
            from the format file.

    Returns:
        List of content from fields in the same order as each field appears in
        the record_format.
    """
    row = []
    degree_formats = ['DDMMXXN', 'DDDMMXXE', 'DDMMSSN', 'DDDMMSSE']
    for index, cell in enumerate(record_format):
        if cell['Start_Column'] == 'na' and 'Concatenate' in cell:
            data = _concat(cell['Concatenate'], row, record_format)
        else:
            data = _slice(line, cell['Start_Column'], cell['Field_Length'])
            if cell['Strip'].lower() == 'y':
                data = data.strip()
            if data.strip() != '':
                try:
                    num_fmt = cell['Number_Format']
                    if cell['Date_Format'] != '':
                        date_object = datetime.strptime(data, cell['Date_Format'])
                        data = date_object.strftime('%Y-%m-%d')
                    elif num_fmt in degree_formats:
                        data = _parse_degrees(data, num_fmt)
                    elif num_fmt == 'i':
                        data = int(data)
                    elif num_fmt[:2] == '.f':
                        data = data.replace(' ', '0')
                        split = int(cell['Field_Length']) - int(num_fmt[2:])
                        data = data[:split] + '.' + data[split:]
                        i = data.find('-')
                        if i != -1:
                            data = data[i:]
                        data = float(data)
                    elif num_fmt != '':
                        raise Exception('Unknown number format: ' + str(num_fmt))
                except Exception as ex:
                    print('Trouble with line:\n' + line)
                    raise
            else:
                data = None  # don't store a bunch of spaces
        row.append(data)
    return row


def _get_record_column(record_format):
    """Return start column and length for record number field in record format.
    """
    for row in record_format:
        if row['Field'].upper().replace(' ', '_') == 'RECORD_NUMBER':
            return row['Start_Column'], row['Field_Length']
    raise Exception('"Record_Number" column not found in record format.')


def parse_file_to_rows(filename, record_formats):
    """Parses the input file into a dictionary of rows.

    Given an NODC archive file and a collection of formats defining how to
    parse records in the file, parse the file and return a dictionary where
    each item represents one data table in the file.

    Args:
        filename: The NODC archive text file to parse.
        record_formats: Dictionary of record formats indexed by record number.

    Returns:
        Dictionary of lists indexed by record number, where each list
        represents rows in the table for that record.
    """
    sample_format = record_formats.itervalues().next()['fields']
    record_start_col, record_field_length = _get_record_column(sample_format)
    rows_by_record = {}
    with open(filename) as f:
        for line in f:
            record_number = _slice(line, record_start_col, record_field_length)
            if record_number in record_formats:
                fmt = record_formats[record_number]
                row = _parse_line(line, fmt['fields'])
                if record_number not in rows_by_record:
                    rows_by_record[record_number] = []
                rows_by_record[record_number].append(row)
            else:
                m = 'Warning: Record format not found for record number - {0}'
                print(m.format(record_number))
    return rows_by_record
