"""Queries itis.gov for taxa information."""

import requests

_base = 'http://www.itis.gov/ITISWebService/jsonservice/'
_solr = 'https://services.itis.gov/'


def _query(uri, params):
    r = requests.get(uri, params=params)
    if r.status_code == 200:
        return r.json()
    else:
        r.raise_for_status()


def fuzzy_search_name(name):
    """Returns TSN for a scientific name using fuzzy search.
    """
    name = name.replace(' ', '\\%20')
    q = 'nameWOInd:' + name + '~0.4'
    params = {'q': q, 'wt': 'json'}
    data = _query(_solr, params)
    if data['response']['numFound'] > 0:
        return data['response']['docs'][0]['tsn']


def exact_search_name(name):
    """Returns TSN for a scientific name if an exact match is found.
    """
    uri = _base + 'searchByScientificName'
    params = {'srchKey': name}
    data = _query(uri, params)
    result = data['scientificNames'][0]
    if result:
        return result['tsn']


def get_tsn_for_sci_name(name):
    """Returns TSN for given scientific name.
    """
    tsn = exact_search_name(name)
    if not tsn:
        tsn = fuzzy_search_name(name)
    return tsn


def get_sci_name_for_tsn(tsn):
    """Returns scientific name for given TSN.
    """
    tsn = int(tsn)
    uri = _base + 'getScientificNameFromTSN'
    params = {'tsn': tsn}
    data = _query(uri, params)
    return data['combinedName']


def get_accepted_info_for_tsn(tsn):
    """Returns accepted TSN and name for given (possibly deprecated) TSN.
    """
    tsn = int(tsn)
    uri = _base + 'getAcceptedNamesFromTSN'
    params = {'tsn': tsn}
    data = _query(uri, params)
    name_dict = data['acceptedNames'][0]
    if name_dict:
        return name_dict['acceptedTsn'], name_dict['acceptedName']
    else:  # name is already accepted
        name = get_sci_name_for_tsn(tsn)
        return tsn, name


def get_accepted_info_for_name(name):
    """Returns accepted TSN and name for (possibly deprecated) scientific name.
    """
    tsn = get_tsn_for_sci_name(name)
    if tsn:
        return get_accepted_info_for_tsn(tsn)
    else:
        return None, None


if __name__ == '__main__':
    # Demonstration
    names = ['Liquidamber styraciflua',
             'Liquidamber',
             'PROACTINOMYCES CITREUS MARINAE',
             'this will not match']
    for name in names:
        t, n = get_accepted_info_for_name(name)
        if n:
            print n
        else:
            print 'no match'
