import csv
import os


def _read_join_by_sequence(format_folder):
    """Imports rules for joining data by sequence number.

    Args:
        format_folder: String indicating folder containing record format files.
            join_by_sequence.csv should also be in this folder.

    Returns:
        List of actions for joining by sequence. Each action is a dictionary
        with these items:
        TargetTable, FromTable, TargetField, SequenceField, MatchField
    """
    filename = os.path.join(format_folder, 'join_by_sequence.csv')
    if not os.path.isfile(filename):
        return []

    rows = []
    with open(filename) as csvfile:
        reader = csv.DictReader(csvfile)
        rows = [row for row in reader]
    return rows


def _read_joins(format_folder):
    """Imports rules for joining data tables.

    Args:
        format_folder: String indicating folder containing record format files.
            joins.csv should also be in this folder.

    Returns:
        List of actions for joining.  Each action is a dictionary with:
        TargetTable, FromTable, TargetField, MatchField
    """
    filename = os.path.join(format_folder, 'joins.csv')
    if not os.path.isfile(filename):
        return []

    rows = []
    with open(filename) as csvfile:
        reader = csv.DictReader(csvfile)
        rows = [row for row in reader]
    return rows


def join_by_sequence(dataframes, format_folder):
    """Appends columns from one DataFrame to another based on sequence number.

    Rows in an archive data file are assigned a sequence number indicating row
    order, with parent items like stations having lower sequence numbers than
    the sample records associated with them. So the game is to find the highest
    sequence number amongst possible parent items that is lower than the
    number for the target child item.
    
    Input dictionary is updated in place with new DataFrames as needed.

    Args:
        dataframes: Dict of pandas DataFrames keyed by record table name.
        format_folder: String indicating folder containing record format files.
            join_by_sequence.csv should also be in this folder.
    """
    join_by_sequence_list = _read_join_by_sequence(format_folder)
    for join in join_by_sequence_list:
        to_df = dataframes[join['TargetTable']]
        from_df = dataframes[join['FromTable']]
        to_seq_idx = to_df.columns.get_loc(join['SequenceField'])
        to_match_idx = to_df.columns.get_loc(join['MatchField'])
        row_count = to_df.shape[0]
        rows = []
        progress = 'Assigning {0}.{1} to row {{0}} of {2}'.format(
            join['TargetTable'], join['TargetField'], row_count)
        for row_idx in range(row_count):
            if (row_idx + 1) % 500 == 0:
                print(progress.format(row_idx + 1))
            row = to_df.iloc[[row_idx]]
            max_seq = to_df.iloc[row_idx, to_seq_idx]
            seq_condition = from_df[join['SequenceField']] <= max_seq
            match_val = to_df.iloc[row_idx, to_match_idx]
            match_condition = from_df[join['MatchField']] == match_val
            matches = from_df.loc[match_condition & seq_condition]
            if matches.shape[0]:
                from_row = matches.ix[matches[join['SequenceField']].idxmax()]
                rows.append(from_row[join['TargetField']])
            else:
                rows.append(None)
        kwargs = {join['TargetField'] : rows}
        dataframes[join['TargetTable']] = to_df.assign(**kwargs)


def join(dataframes, format_folder):
    """Appends columns from one DataFrame to another based on matching fields.

    Input dictionary is updated in place with new DataFrames as needed.

    Args:
        dataframes: Dict of pandas DataFrames keyed by record table name.
        format_folder: String indicating folder containing record format files.
            joins.csv should also be in this folder.
    """
    join_list = _read_joins(format_folder)
    m = 'Joining {0}.{1} to {2}'
    for join in join_list:
        print(m.format(
            join['FromTable'], join['TargetField'], join['TargetTable']))
        to_df = dataframes[join['TargetTable']]
        from_flds = [join['MatchField'], join['TargetField']]
        from_df = dataframes[join['FromTable']][from_flds]
        key = join['MatchField']
        df = to_df.join(from_df.set_index(key), on=key)
        dataframes[join['TargetTable']] = df
