from collections import defaultdict
import csv
from datetime import datetime
import math
import os
from time import sleep

import pandas as pd

from codes import codes_to_dataframes
from formats import read_record_formats, parse_file_to_rows
from joins import join_by_sequence, join
from itis import get_accepted_info_for_tsn as query_tsn
from taxa import read_taxnum_into_dataframe as read_taxnum


def _cols_for_format(record_format_fields):
    return [r['Field'] for r in record_format_fields]


def _merge_duplicate_columns(dataframe):
    headers = dataframe.dtypes.index
    name_counts = defaultdict(int)
    for h in headers:
        name_counts[h] += 1
    drop_columns = []
    for col, count in name_counts.iteritems():
        if count > 1:
            # Overwrite first instance of column with total of values
            total = dataframe[col].sum(axis=1)
            indices = [i for i, x in enumerate(dataframe.columns.get_loc(col))
                       if x == True]
            dataframe.iloc[:, indices[0]] = total
            drop_columns.extend(indices[1:])
    if drop_columns:
        keep_columns = [j for j, c in enumerate(dataframe.columns)
                        if j not in drop_columns]
        dataframe = dataframe.iloc[:, keep_columns]
    return dataframe


def _assign_codes(dataframe, code_dfs, record_format):
    ignore_codes = ['', 'taxa']
    out_columns = []
    for i, field in enumerate(record_format['fields']):
        out_columns.append(field['Field'])
        code_table_name = field['Code']
        if code_table_name not in ignore_codes:
            if code_table_name not in code_dfs:
                m = 'Warning: Code table not loaded for ' + code_table_name
                print(m)
            else:
                code_df = code_dfs[code_table_name]
                dataframe = pd.merge(dataframe, code_df, how='left',
                                     left_on=[field['Field']],
                                     right_on=['Code'])
                name = field['Field'] + '_Description'
                dataframe.rename(columns={'Code_Meaning': name}, inplace=True)
                dataframe.drop('Code', axis=1, inplace=True)
                out_columns.append(name)
    return dataframe[out_columns]


def _query_itis(dataframe):
    max_attempts = 5
    
    if 'TSN1' not in dataframe:
        return dataframe

    tsns = [int(t) for t in dataframe.TSN1.unique() if t and not math.isnan(t)]
    if 'TSN2' in dataframe:
        tsns2 = [int(t) for t in dataframe.TSN2.unique()
                 if t and not math.isnan(t)]
        has_tsn2 = len(tsns2)
        if has_tsn2:
            tsns = list(set(tsns + tsns2))
    if not tsns:
        return dataframe

    rows = []
    m = 'Max attempts reached. Could not retrieve ITIS data for TSN {0}'
    for tsn in tsns:
        accepted_tsn, name = None, None
        for attempt in range(max_attempts):
            try:
                print('Querying ITIS for ' + str(tsn))
                accepted_tsn, name = query_tsn(tsn)
                break
            except:
                if attempt == max_attempts - 1:
                    print(m.format(tsn))
                sleep(0.5)  # Give ITIS time to wake up
        if name:
            rows.append([tsn, name])

    itis_cols = ['TsnKey', 'ItisName']
    names = pd.DataFrame(rows, columns=itis_cols)
    df = pd.merge(dataframe, names, how='left',
                   left_on=['TSN1'], right_on=['TsnKey'])
    df['ITIS_Name1'] = df['ItisName']
    df.drop(itis_cols, axis=1, inplace=True)
    if has_tsn2:
        df2 = pd.merge(df, names, how='left',
                       left_on=['TSN2'], right_on=['TsnKey'])
        df2['ITIS_Name2'] = df2['ItisName']
        df2.drop(itis_cols, axis=1, inplace=True)
        df = df2
    return df

    
def _assign_taxa(dataframe, taxa, record_format, use_itis):
    """Assumes only one taxonomy column per file"""
    out_columns = []
    has_taxa = False
    for column in dataframe:
        out_columns.append(column)
        matches = [f for f in record_format['fields']
                   if f['Code'].lower() == 'taxa' and f['Field'] == column]
        if matches:
            dataframe = pd.merge(dataframe, taxa, how='left',
                                 left_on=[column],
                                 right_on=['Taxon_Code'])
            out_columns.extend(list(taxa)[1:])
            dataframe.drop('Taxon_Code', axis=1, inplace=True)
            has_taxa = True
    if has_taxa:
        df = dataframe[out_columns]
        if use_itis:
            df = _query_itis(df)
    else:
        df = dataframe
    return df


def parse_file(filename, format_folder, code_dfs, taxa, use_itis):
    """Parses the input file into a collection of pandas DataFrames.

    Given an NODC archive file and a collection of formats defining how to
    parse records in the file, parse the file and return a collection of
    pandas DataFrames representing the data tables in the file.

    Args:
        filename: The NODC archive text file to parse.
        format_folder: Folder with record format files.
        code_dfs: Dictionary of pandas DataFrames of NODC codes indexed by code
            table identifier.
        taxa: Pandas DataFrame of Taxon_Code, Taxon_Name1, TSN1, Taxon_Name2,
            TSN2.
        use_itis: Boolean indicating whether ITIS should be queried for current
            taxa names.

    Returns:
        Dictionary of pandas DataFrames indexed by record name.
    """
    record_formats = read_record_formats(format_folder)
    rows_by_record = parse_file_to_rows(filename, record_formats)

    dataframes = {}
    for record_number, rows in rows_by_record.iteritems():
        fmt = record_formats[record_number]
        df = pd.DataFrame(rows, columns=_cols_for_format(fmt['fields']))
        df = _assign_codes(df, code_dfs, fmt)
        df = _assign_taxa(df, taxa, fmt, use_itis)
        df.name = fmt['name']
        dataframes[fmt['name']] = df

    # Sometimes they have separate columns for, e.g., long vs. short hauls
    # Combine them so databases don't have trouble with duplicate column names
    for name, df in dataframes.iteritems():
        dataframes[name] = _merge_duplicate_columns(df)
    
    join_by_sequence(dataframes, format_folder)
    join(dataframes, format_folder)
    return dataframes


def extract_to_csv(filename, format_folder, code_tables, taxa, use_itis,
                   out_folder):
    """Extracts data from an NODC archive file into CSV files.

    Given an NODC archive file, extract data for each record in the file into
    its own CSV file. A record refers to file header, station header, site
    header, etc.  Output files are named according to record type.

    Args:
        filename: The NODC archive text file to split.
        format_folder: Folder with record format files.
        code_tables: Dictionary of NODC codes indexed by code table identifier.
        taxa: Pandas DataFrame of Taxon_Code, Taxon_Name, and Taxon_Level.
        use_itis: Boolean indicating whether ITIS should be queried for current
            taxa names.
        out_folder: Output folder where each record file will be written.
    """
    dataframes = parse_file(filename, format_folder, code_tables, taxa,
                            use_itis)
    for name, df in dataframes.iteritems():
        out_file = os.path.join(out_folder, name + '.csv')
        df.to_csv(out_file, index=False)


def _slice(line, start_column, field_length):
    """Extracts a slice from a line of text with a one-based index.

    Slices the input line of text to extract a portion defined by a start
    column and the length of the cell.  We assume NODC column definitions,
    which are one-based, e.g., the first column is column 1.

    Args:
        line: The string from which a portion is to be extracted.
        start_column: One-based start index of the cell. 
        field_length: Number of characters in the cell.

    Returns:
        The extracted string.
    """
    python_start_index = int(start_column) - 1
    python_end_index = python_start_index + int(field_length)
    return line[python_start_index:python_end_index]


def split_file(filename, record_start_col, record_field_length, out_folder):
    """Splits an NODC archive file into separate files per record.

    Given an NODC archive file, split each record in the file into its own
    file. A record refers to file header, station header, site header, etc.
    Output files are named according to record identifier.

    Args:
        filename: The NODC archive text file to split.
        record_start_col: One-based start index of record identifier. 
        record_field_length: Number of characters in the record identifier.
        out_folder: Output folder where each record file will be written.
    """
    lines_by_record = {}
    with open(filename) as f:
        for line in f:
            record_number = _slice(line, record_start_col, record_field_length)
            if record_number not in lines_by_record:
                lines_by_record[record_number] = []
            lines = lines_by_record[record_number]
            lines.append(line)
    for record, lines in lines_by_record.iteritems():
        out_file = os.path.join(out_folder, record + '.txt')
        with open(out_file, 'wb') as f:
            f.write(''.join(lines))


def main(data_file, format_folder, code_file, taxnum_file, out_folder,
         use_itis=True):
    print('Processing ' + data_file)
    print('    Reading taxa tables')
    taxa = read_taxnum(taxnum_file)

    print('    Reading code tables')
    code_dfs = codes_to_dataframes(code_file)

    print('    Parsing ' + data_file)
    extract_to_csv(data_file, format_folder, code_dfs, taxa, use_itis,
                   out_folder)
