"""Reads NODC taxonomic codes."""

import pandas as pd


def _resolve_deprecated(taxa, deprecated):
    """Updates taxa dictionary with current descriptions for deprecated items.

    Updates taxa dictionary with current descriptions for deprecated items.
    Deprecated items include the code for the currently valid version of the
    taxon.  An entry for the deprecated item is added to the taxa dictionary
    by pointing to the values for the currently valid version from the taxa
    dictionary.  The input taxa dictionary is updated in place.

    Args:
        taxa: Dictionary keyed by valid taxonomy code, with a list of valid
            taxonomic descriptions as its value.
        deprecated: Dictionary of deprecated taxonomy codes, with a list of
            the valid codes as its value.
    """

    while len(deprecated):
        tmp = {}
        previous_count = 0
        redo_count = 0
        for old_code, use_codes in deprecated.iteritems():
            for code in use_codes:
                previous_count += 1
                if code in taxa:
                    if old_code in taxa:
                        items = taxa[old_code]
                    else:
                        items = []
                        taxa[old_code] = items
                    items.extend(taxa[code])
                else:
                    # Sometimes the provided current code is also deprecated, so we
                    # must store them for now and then iterate.
                    if old_code in tmp:
                        redo = tmp[old_code]
                    else:
                        redo = []
                        tmp[old_code] = redo
                    redo.extend([code])
                    redo_count += 1
        if previous_count == redo_count:
            # We're not getting any smaller
            for old_code in tmp:
                if old_code not in taxa:
                    taxa[old_code] = []
            if redo_count:
                s = 's' if redo_count > 1 else ''
                m = 'Could not resolve the following deprecated code' + s + ':'
                parts = '\n'.join(sorted([str(code) for code in tmp]))
                #print('Warning -- ' + m + '\n' + parts)
            break  
        deprecated = tmp


def _parse_use_codes(txt):
    """Get more current taxon codes for a deprecated code

    Deprecated codes have text indicating which code or codes to use instead.

    Args:
        txt: Text potentially containing USE codes. Examples:
            !USE 16090209 OR WAS 16090202!
            !USE 0402020408  !
            !USE 16080504 OR 16080503!
            !USE  -NEVER-    !

    Returns:
        list of integer taxonomy codes
    """
    txt = txt.replace('!', '').split(' OR WAS ')[0].replace('USE ', '').strip()
    codes = [int(c) for c in txt.split(' OR ') if 'NEVER' not in c]
    return codes


def _pick_matching_code(code_to_match, candidates):
    """Given a continued taxonomy code, find the matching code and data.

    The matching code is found by taking the difference of code numbers. The
    smallest difference wins.  This assumes that the original authors of the
    taxonomy code wanted to keep continuation codes as similar as possible to
    the main codes.

    Args:
        code_to_match: Taxonomy code for which a match should be found.
        candidates: List of possible matching codes.

    Returns:
        The best matching taxonomy code.
    """
    code = int(code_to_match)
    diff = [abs(code - int(c)) for c in candidates]
    index = diff.index(min(diff))
    return candidates[index]

        
def _resolve_continued(taxa, codes_by_name, continued):
    """Updates taxa dictionary with information for continued codes.

    This function copies taxa information from main records to continuation
    records, updating the taxa dictionary in place.
    
    Taxa codes are generally hierarchical. For example, 64650105 may be a genus
    level code, and species under that genus may have codes 6465010501,
    6465010502, 6465010503, and so on. But if after 6465010599, more species
    under that genus exist, then additional species must to continue under a
    different set of beginning digits, such as 64650106, e.g., 6465010601,
    6465010602.  When this happens, a record is written that indicates 64650106
    is a continuation of an existing taxon. However, the record does NOT
    include the TSN or hierarchy level.

    The continuation record also does not include the main code of the taxon it
    is continuing.  One can locate a candidate main record by matching the
    taxon name; however, more than one taxon may have the same name. Therefore,
    all possible matches must be compared to find the best matching code.

    Args:
        taxa: dict keyed by integer taxonomy code, where each item is a list of
            one or more dictionaries. Each dict in the list has these items:
            code, name, level, tsn
        codes_by_name: dict keyed by taxon name, where each item is a list of
            taxonomy codes with that name.
        continued: dict keyed by integer taxonomy code, where each item is the
            taxon name of a continued taxon entry.  Main taxon entries that
            match items in this dict will be located and added to the input
            taxa dict.
    """
    for cont_code, name in continued.iteritems():
        if cont_code not in taxa:
            if name in codes_by_name:
                candidates = codes_by_name[name]
                if len(candidates) == 1:
                    taxa[cont_code] = taxa[candidates[0]]
                else:
                    winner = _pick_matching_code(cont_code, candidates)
                    taxa[cont_code] = taxa[winner]
            else:
                taxa[cont_code] = []

    
def read_taxnum_as_dict(taxnum_file):
    """Reads taxonomy codes into a dictionary.

    Reads taxonomy codes into a dictionary.  Note that a code may be
    deprecated with up to two more current taxon names and TSNs.  If a code
    is deprecated, the taxon names and TSNs are those of the more current
    codes rather than the deprecated code.

    Args:
        taxnum_file: Path and filename of taxa code file, TAXNUM.DAT

    Returns:
        dict keyed by integer taxonomy code, where each item is a list of one
        or more dictionaries. Each dict in the list has these items:
        code, name, level, tsn
    """
    levels = {'K': 'Kingdom',
              'P': 'Phylum',
              'C': 'Class',
              'O': 'Order',
              'F': 'Family',
              'G': 'Genus',
              'S': 'Species',
              'V': 'Subspecies or variety',
              ' ': ''}
    taxa = {}
    deprecated = {}
    codes_by_name = {}
    continued = {}
    
    with open(taxnum_file) as f:
        lines = f.read().splitlines()
    is_v8 = len(lines[0]) > 102
    for line in lines:
        name = line[0:45].strip()
        code = int(line[80:92].strip())
        control_code = line[93]
        if control_code == ' ':
            if is_v8 and line[104] != '.':
                tsn = int(line[99:105])
            else:
                tsn = ''
            level = levels[line[78]]
            taxa[code] = [{'code': code,
                           'name': name,
                           'level': level,
                           'tsn': tsn}]
            if name in codes_by_name:
                codes = codes_by_name[name]
            else:
                codes = []
                codes_by_name[name] = codes
            codes.append(code)
        elif control_code == 'W' and code not in deprecated:
            codes = _parse_use_codes(name)
            if codes:
                deprecated[code] = codes
        elif control_code == 'K':
            name = name.split(' *')[0]
            if code in continued and continued[code] != name:
                m = 'Different names found for continuation code ' + code
                raise Exception(m)
            continued[code] = name

    _resolve_continued(taxa, codes_by_name, continued)
    _resolve_deprecated(taxa, deprecated)
                
    return taxa


def read_taxnum_into_dataframe(taxnum_file):
    """Reads taxonomy codes into a pandas DataFrame.

    Reads taxonomy codes into a pandas DataFrame.  Note that a code may be
    deprecated with up to two more current taxon names and TSNs.  If a code
    is deprecated, the taxon names and TSNs are those of the more current
    codes rather than the deprecated code.

    Args:
        taxnum_file: Path and filename of taxa code file, TAXNUM.DAT

    Returns:
        DataFrame with columns for Taxon_Code, Taxon_Name1, TSN1, ITIS_Name1,
        Taxon_Name2, TSN2, and ITIS_Name2. ITIS_Name1 and ITIS_Name2 are
        placeholders and are not populated by this function.
    """

    taxa = read_taxnum_as_dict(taxnum_file)
    columns = ['Taxon_Code', 'Taxon_Name1', 'TSN1', 'ITIS_Name1',
               'Taxon_Name2', 'TSN2', 'ITIS_Name2']
    rows = []
    for code, items in taxa.iteritems():
        row = [code]
        for item in items:
            row.extend([item['name'], item['tsn']])
        row = row + [None] * (7 - len(row))
        rows.append(row)
    df = pd.DataFrame(rows, columns=columns)
    df.name = 'Taxa_Database'
    return df
